#!/usr/bin/perl -w
#
# add.cgi
# John Simpson <jms1@jms1.net>
#
# Add a rule
#
# $Id: add.cgi 106 2012-07-21 20:39:17Z jms1 $
#
###############################################################################
#
# Copyright (C) 2012 John M. Simpson.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 or version 3 of the
# license, at your option.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

require 5.003 ;
use strict ;
use CGI qw ( :standard ) ;

use Rules ;

###############################################################################
###############################################################################
###############################################################################

require_login() ;

########################################
# Check for cancel button.

if ( param ( 'cancel' ) )
{
	logit "cancelled\n" ;
	redirect_home() ;
}

########################################
# Read form data.

my $p_phase	= ( param ( 'phase'  ) || '' ) ;
my $p_seq	= ( param ( 'seq'    ) || 0  ) ;
my $p_go	= ( param ( 'go'     ) || '' ) ;
my $p_type	= ( param ( 'type'   ) || '' ) ;
my $p_sender	= ( param ( 'sender' ) || '' ) ;
my $p_ip	= ( param ( 'ip'     ) || '' ) ;
my $p_rbl	= ( param ( 'rbl'    ) || '' ) ;
my $p_target	= ( param ( 'target' ) || '' ) ;
my $p_delay	= ( param ( 'delay'  ) || '' ) ;
my $p_accept	= ( param ( 'accept' ) || 0  ) ? 1 : 0 ;
my $p_desc	= ( param ( 'desc'   ) || '' ) ;

########################################
# Basic sanity check

unless ( ( $p_phase =~ m|^\d+$| ) && ( $p_seq =~ m|^\d+| ) )
{
	logit "ERROR: invalid phase [$p_phase] or seq [$p_seq]\n" ;
	redirect_home() ;
}

########################################
# Make sure the user is authorized to work with rules in this phase.

my $query = "SELECT level FROM phases WHERE phase = $p_phase" ;
my $plevel = SQL_select ( $query ) ;
$plevel ||= 0 ;

if ( $access < $plevel )
{
	logit "ERROR: attempt to edit rule without proper access: phase=[$p_phase] seq=[$p_seq] plevel=[$plevel] access=[$access] userid=[$userid] focus[$focus]\n" ;
	redirect_home() ;
}

########################################
# Are we actually doing anything?

if ( $p_go )
{
	########################################
	# Move any existing rules out of the way.

	SQL_transaction() ;

	$query = 'SELECT id FROM rules'
		. ' WHERE ' . SQL_quote ( $focus ) . ' LIKE recipient'
		. " AND phase = $p_phase AND seq > $p_seq"
		. ' ORDER BY seq DESC' ;
	( my $id ) = SQL_selectfirst ( $query ) ;
	while ( $id )
	{
		my $q2 = "UPDATE rules SET seq = seq + 1 WHERE id = $id" ;
		SQL_do ( $q2 ) ;

		( $id ) = SQL_selectnext() ;
	}

	SQL_commit() ;

	########################################
	# Build the INSERT query

	my $recip = $focus ;
	$recip =~ s|^\@$|%| ;
	$recip =~ s|^(\@.+)|%$1| ;

	my @flist = ( 'phase' , 'seq' , 'recipient' , 'type' , 'description' ) ;
	my @dlist = ( $p_phase , $p_seq + 1 , SQL_quote ( $recip ) ,
		SQL_quote ( $p_type ) , SQL_quote ( $p_desc ) ) ;

	if ( $p_type eq 'A' )
	{
		# The only thing an all-messages rule needs is "type='A'",
		# which has already been done.
	}
	elsif ( $p_type eq 'U' )
	{
		# The only thing an aUth rule needs is "type='U'",
		# which has already been done.
	}
	elsif ( $p_type eq 'E' )
	{
		unless ( validate_regex ( $p_sender ) )
		{
			bad "Invalid regular expression \"$p_sender\"." ;
		}
		push ( @flist , 'sender' ) ;
		push ( @dlist , SQL_quote ( $p_sender ) ) ;
	}
	elsif ( $p_type eq 'I' )
	{
		unless ( validate_ip ( $p_ip ) )
		{
			bad "Invalid IP or CIDR address \"$p_ip\"." ;
		}
		push ( @flist , 'ip' ) ;
		push ( @dlist , SQL_quote ( $p_ip ) ) ;
	}
	elsif ( $p_type eq 'R' )
	{
		unless ( validate_rbl ( $p_rbl ) )
		{
			bad "Invalid RBL zone address \"$p_rbl\"." ;
		}
		push ( @flist , 'rbl' ) ;
		push ( @dlist , SQL_quote ( $p_rbl ) ) ;
	}
	elsif ( $p_type eq 'T' )
	{
		unless ( validate_regex ( $p_target ) )
		{
			bad "Invalid regular expression \"$p_target\"." ;
		}
		push ( @flist , 'target' ) ;
		push ( @dlist , SQL_quote ( $p_target ) ) ;
	}
	elsif ( $p_type eq 'G' )
	{
		unless ( validate_grey ( $p_delay ) )
		{
			bad "Invalid greylist delay \"$p_delay\"." ;
		}
		push ( @flist , 'delay' ) ;
		push ( @dlist , $p_delay ) ;
	}
	elsif ( $p_type eq 'D' )
	{
		# The only thing the debug rule needs is "type='D'",
		# which has already been done.
	}
	else
	{
		fatal "Invalid rule type \"$p_type\"" ;
	}

	push ( @flist , 'accept' ) ;
	push ( @dlist , $p_accept ? SQL_true() : SQL_false() ) ;

	my $query = 'INSERT INTO rules ( ' . join ( ' , ' , @flist )
		. ' ) VALUES ( ' . join ( ' , ' , @dlist ) . ' )' ;

	########################################
	# Do the deed.

	SQL_do ( $query ) ;
	redirect_home() ;
}

########################################
# Not adding a rule yet. Show the form.

my $showu = ( $ENV{'SHOW_AUTH_RULES'} || 0 ) ;

my $nrows = 0 ;
for my $k ( keys %typename )
{
	$nrows += ( ( $k eq 'U' ) ? ( $showu ? 1 : 0 ) : 1 ) ;
}

my $ustart = <<EOF ;
<!-- This rule type is only available with the combined patch v7 or later.
     http://qmail.jms1.net/patches/combined.shtml
     Please set SHOW_AUTH_RULES in your .htaccess file if necessary.

EOF

my $uend   = ' -->' ;

if ( $showu )
{
	$ustart = $uend = '' ;
}

my $recip = focus_string() ;

print page_start ( 'Add a rule' ) ;
print parse_template ( 'add-rule.txt' ,
	'phase'  => $p_phase ,
	'seq'    => $p_seq   ,
	'recip'  => $recip   ,
	'nrows'  => $nrows   ,
	'ustart' => $ustart  ,
	'uend'   => $uend    ) ;
print page_end() ;
exit 0 ;
