-- Create postgresql tables for grey4 database
-- John Simpson <jms1@jms1.net> 2011-06-13
--
-- $Id: create-tables.pgsql 119 2012-07-22 22:43:42Z jms1 $
--
-- NOTE: Before you run this, you must ensure that your postgresql server
-- has userids for "g4web and "g4mail". The "g4web" userid will be used
-- by the web interface to edit the data, while the "g4mail" userid will
-- be used by qmaild to query the rules and add log entries.
--
-- If you are using different userids, you will need to edit this file and
-- change the userids at the end of every GRANT statement.
--
------------------------------------------------------------------------------
--
-- Drop any existing tables.

DROP TABLE sessions ;
DROP TABLE log ;
DROP TABLE greylist ;
DROP TABLE rules ;
DROP TABLE phases ;

-- Create "phases" table and set security.

CREATE TABLE phases
(
	phase		SMALLINT	NOT NULL PRIMARY KEY ,

	-- level needed to edit rules in this phase
	-- 1=user, 2=domain admin, 3=machine admin
	level		SMALLINT	NOT NULL ,

	description	VARCHAR(80)	NOT NULL
) ;

GRANT SELECT ON phases TO g4mail ;
GRANT SELECT,INSERT,UPDATE,DELETE ON phases TO g4web ;

-- Create "rules" table and set security.

CREATE TABLE rules
(
	-- Unique identifier

	id		SERIAL		NOT NULL PRIMARY KEY ,

	-- Phase/sequence

	phase		SMALLINT	NOT NULL REFERENCES phases ( phase ) ,
	seq		INTEGER		NOT NULL ,

	-- Recipient (owner) will be one of:
	-- * full email address, rule affects just that user's mail
	-- * domain, affects mail sent to all addresses in the domain ("%@domain")
	-- * system-wide, affects all incoming mail on the machine ("%")

	recipient	VARCHAR(80)	NOT NULL ,

	-- Type code selects which incoming messages, based on sender info:
	-- * E	Email: This rule applies to all messages whose envelope sender
	--	(the MAIL FROM argument) matches the regular expression in the
	--	"sender" field.
	-- * I	IP: This rule applies to all messages where the sending IP
	--	(the SMTP client) matches the CIDR address in the "ip" field.
	-- * R	RBL: This rule applies to all messages where the sending IP
	--	(the SMTP client) is found in the DNS-based RBL zone listed
	--	in the "rbl" field.
	-- * T	Target: This rule match all messages sent TO the address in
	--	the "target" field.
	-- * U	aUth: This rule matches all messages sent by a client who has
	--	sent a successful AUTH command.
	-- * A	All: This rule applies to all messages.
	-- Other "type" values:
	-- * G	If the message has not already been accepted or rejected,
	--	check the sending (IP,sender,recipient) against a greylist.
	--	The "delay" field tells how long each sender must wait.
	-- * D	If the message has not already been accepted or rejected,
	--	enable debugging of the message processing. This can be used
	--	to assist in troubleshooting why some messages are being
	--	accepted or rejected, but should not be used all the time
	--	because it will make your log files grow, and because it can
	--	be a violation of your users' privacy.

	type		CHAR(1)		NOT NULL ,
	sender		VARCHAR(250)	,
	ip		CHAR(18)	,
	rbl		VARCHAR(250)	,
	delay		INTEGER		,
	target		VARCHAR(250)	,

	-- accept: If the message matches the criteria expressed in an [AEIR]
	--	rule, the message will be accepted (if true) or rejected (if
	--	false.) The "G" and "D" rules do not use this field.

	accept		BOOL		NOT NULL ,

	-- description

	description	VARCHAR(80)	,

	-- statistics

	created		TIMESTAMP WITH TIME ZONE
					NOT NULL DEFAULT NOW()
) ;

CREATE UNIQUE INDEX rules_prs ON rules ( phase , recipient , seq ) ;

GRANT SELECT                      ON rules        TO g4mail ;
GRANT SELECT,INSERT,UPDATE,DELETE ON rules        TO g4web ;
GRANT               UPDATE        ON rules_id_seq TO g4web ;

-- Create the greylist table.

CREATE TABLE greylist
(
	id		SERIAL		NOT NULL UNIQUE ,

	sender		VARCHAR(250)	NOT NULL ,
	recipient	VARCHAR(250)	NOT NULL ,
	ip		CHAR(18)	NOT NULL ,

	first		INTEGER		NOT NULL ,
	last		INTEGER		NOT NULL ,

	count_pre	INTEGER		NOT NULL DEFAULT 1 ,
	count_msg	INTEGER		NOT NULL DEFAULT 0 ,

	PRIMARY KEY ( sender , recipient , ip )
) ;

GRANT SELECT,DELETE ON greylist TO g4web ;

GRANT INSERT,SELECT,UPDATE ON greylist        TO g4mail ;
GRANT        SELECT,UPDATE ON greylist_id_seq TO g4mail ;

-- Create the log table.

CREATE TABLE log
(
	id		SERIAL				PRIMARY KEY ,
	ruleid		INTEGER				NOT NULL ,

	dt		TIMESTAMP WITH TIME ZONE	NOT NULL DEFAULT NOW() ,
	sender		VARCHAR(250)			NOT NULL ,
	recipient	VARCHAR(250)			NOT NULL ,
	ip		CHAR(15)			NOT NULL ,

	disp		VARCHAR(20)			NOT NULL
) ;

GRANT SELECT ON log TO g4web ;
GRANT INSERT ON log TO g4mail ;

GRANT SELECT,UPDATE ON log_id_seq TO g4mail ;

-- Create the sessions table.

CREATE TABLE sessions
(
	-- session id is md5_base64() output, always 22 bytes.
	id		CHAR(22)			PRIMARY KEY ,

	-- only this IP is allowed to use the session.
	ip		CHAR(15)			NOT NULL ,

	-- userid and access level granted with the session.
	userid		VARCHAR(80)			NOT NULL ,
	level		SMALLINT			NOT NULL ,

	-- session tracking.
	first		INTEGER				NOT NULL ,
	expires		INTEGER				NOT NULL ,
	hits		INTEGER				NOT NULL DEFAULT 1 ,
	ended		BOOLEAN				NOT NULL DEFAULT FALSE
) ;

GRANT SELECT,INSERT,UPDATE,DELETE ON sessions TO g4web ;
