#!/usr/bin/perl -w
#
# report.pl
# John Simpson <jms1@jms1.net> 2012-07-04
#
# This is a command-line tool which allows a machine owner to produce a list
# of the rules in effect for:
# - a single mailbox
# - all mailboxes in a domain
# - all domains and mailboxes on the entire server
#
# Note that it does use the "Rules.pm" module, however it doesn't try to find
# the database connection information using environment variables. (This only
# works for CGI scripts, because Apache adds them to the environment before
# executing the script.) If you want to use this script, you will need to add
# or change the $dbconn, $dbuser, and $dbpass values below the "use Rules"
# line.
#
# $Id: report.pl 112 2012-07-22 01:06:20Z jms1 $
#
###############################################################################
#
# Copyright (C) 2012 John M. Simpson.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 or version 3 of the
# license, at your option.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

require 5.003 ;
use strict ;

use Rules ;

########################################
# CONFIGURE THE DATABASE HERE

$dbconn = 'dbi:Pg(PrintError=>0):dbname=rules' ;
$dbuser = 'g4web' ;
$dbpass = 'web' ;

###############################################################################
#
# usage

sub usage()
{
	print <<EOF ;
$0 [target]

Show the rules which apply to incoming mail for a given target. The target
can be one of the following:

user\@domain	One specific mailbox. This will include all "user-*\@domain"
		addresses as well.

domain		All rules for a specific domain. This will include any
		per-mailbox rules which may have been set up by users.

(nothing)	Only show the system-wide rules.

-h		Show this help message.

EOF
	exit 0 ;
}

###############################################################################
#
# padding function

sub pad($$$)
{
	my $string = shift ;
	my $pad	   = shift ;
	my $len    = shift ;

	if ( length ( $string ) >= $len )
	{
		return $string ;
	}

	my $rv = $string . ( $pad x ( $len - length ( $string ) ) ) ;

	return $rv ;
}

###############################################################################
###############################################################################
###############################################################################
#
# figure out what we're being asked for

my $focus = ( shift @ARGV || '' ) ;
if ( $focus eq '-h' )
{
	usage() ;
}

###############################################################################
#
# get phase names

my $query = 'SELECT phase , description FROM phases' ;
my %pdesc = SQL_select ( $query ) ;

###############################################################################
#
# figure out what we're looking for

my $where = "WHERE type != 'D'" ;
my $title = "Rules for mail sent to " ;

if ( $focus =~ m|^(.+)\@(.+)| )		# user@domain
{
	my ( $user , $dom ) = ( $1 , $2 ) ;
	$user =~ s|\-.*$|| ;
	$focus = "$user\@$dom" ;

	$where .= ' AND ' . SQL_quote ( $focus ) . ' LIKE recipient' ;
	$title .= $focus ;
}
elsif ( $focus =~ m|^\@?(.+)$| )	# @domain or domain
{
	my $dom = $1 ;

	$where .= ' AND ( recipient LIKE ' . SQL_quote ( "\%\@$dom" )
		. ' OR recipient = ' . SQL_quote ( '%' ) . ' )' ;
	$title .= "domain $dom" ;
}
elsif ( $focus eq '' )			# no arg, system-wide
{
	$title .= "this server" ;
}
else
{
	print "ERROR: \"$focus\" does not look valid\n\n" ;
	usage() ;
}

###############################################################################
#
# get the rules, build @output

$query = <<EOF ;
SELECT id , recipient , phase , seq , type , accept ,
	sender , ip , rbl , delay , target , description
FROM rules $where
ORDER BY phase , recipient , seq
EOF

my @data = SQL_select ( $query ) ;
my @output = ( "Seq\tDisp\tType\tValue\tDescription" ) ;
my $lphase = 0 ;
my $lrecipient = '' ;

while ( $#data > -1 )
{
	my $id		=   shift @data ;
	my $recipient	=   shift @data ;
	my $phase	=   shift @data ;
	my $seq		=   shift @data ;
	my $type	=   shift @data ;
	my $accept	=   shift @data ;
	my $sender	= ( shift @data || '' ) ;
	my $ip		= ( shift @data || '' ) ;
	my $rbl		= ( shift @data || '' ) ;
	my $delay	= ( shift @data || 0  ) ;
	my $target	= ( shift @data || '' ) ;
	my $description	= ( shift @data || ' ' ) ;

	if ( $phase != $lphase )
	{
		push ( @output , "[$phase] $pdesc{$phase} " ) ;
		$lphase = $phase ;
	}

	if ( $recipient ne $lrecipient )
	{
		my $x = $recipient ;
		$x =~ s|^\%|(all)| ;
		push ( @output , "  TO: $x " ) ;

		$lrecipient = $recipient ;
	}

	my $value = '' ;
	   if ( $type eq 'E' ) { $value = $sender ;	}
	elsif ( $type eq 'I' ) { $value = $ip ;		}
	elsif ( $type eq 'R' ) { $value = $rbl ;	}
	elsif ( $type eq 'T' ) { $value = $target ;	}
	elsif ( $type eq 'G' ) { $value = $delay ;	}

	my $alabel = ( $accept ? 'ACCEPT' : 'REJECT' ) ;
	if ( $type eq 'G' ) { $alabel = 'DELAY' ; }

	my $zseq = ( rule_is_final($type) ? '*** ' : '' ) . $seq ;

	my $ztype = ( $typename{$type} || "($type)" ) ;

	push ( @output , "$zseq\t$alabel\t$ztype\t$value\t$description" ) ;
}

###############################################################################
#
# figure out column widths

my @cwid = () ;

for my $line ( @output )
{
	next unless ( $line =~ m|\t| ) ;

	my @w = split ( m|\t| , $line ) ;
	for my $n ( 0 .. $#w )
	{
		if ( length ( $w[$n] ) > ( $cwid[$n] || 0 ) )
		{
			$cwid[$n] = length ( $w[$n] ) ;
		}
	}
}

###############################################################################
#
# generate output

my $format = sprintf ( "  %%%ds %%-%ds %%-%ds %%-%ds %%-%ds\n" , @cwid ) ;
my $lwid = length ( sprintf ( $format , '' , '' , '' , '' , '' ) ) ;

print "$title\n\n" ;

for my $line ( @output )
{
	if ( $line =~ m|^\[| )
	{
		print pad ( $line , '=' , $lwid + 2 ) , "\n" ;
		next ;
	}
	elsif ( $line !~ m|\t| )
	{
		print pad ( $line , '-' , $lwid ) , "\n" ;
		next ;
	}

	printf $format , split ( m|\t| , $line ) ;
}

