#!/usr/bin/perl -w
#
# rcptcheck.pl
# John Simpson <jms1@jms1.net>
#
# $Id: rcptcheck.pl 127 2012-07-24 02:58:15Z jms1 $
#
# Designed to be called by qmail-smtpd when patched with the RCPTCHECK patch,
# either by itself, or as part of version 7.07 or later of my qmail combined
# patch.
#
# The original RCPTCHECK patch info:
#	http://www.soffian.org/downloads/qmail/qmail-smtpd-doc.html
#
# My combined patch:
#	http://qmail.jms1.net/patches/combined-details.shtml#7.07
#
###############################################################################
#
# Copyright (C) 2008,2012 John M. Simpson.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 or version 3 of the
# license, at your option.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

require 5.003 ;
use strict ;

use DBI ;
use Net::DNS ;

###############################################################################
#
# Global variables

my $log_rules		= 0 ;
my $grey_match_24	= 0 ;

my ( $sql_server , $sql_userid , $sql_passwd , 
	$dbh , $sender , $recipient , $original_recipient , $ip ,
	$ruleid , $phase , $seq ) ;

###############################################################################
#
# SQL functions

sub sql_connect()
{
	$dbh ||= DBI->connect ( $sql_server , $sql_userid , $sql_passwd )
		or bad ( 120 , "unable to connect to SQL server: "
			. $DBI::errstr ) ;
}

sub sql_disconnect()
{
	return unless $dbh ;
	$dbh->disconnect() ;
	undef $dbh ;
}

END { sql_disconnect() } ;

sub sql_select($)
{
	my $query = shift ;

	sql_connect() ;

	my $sth = $dbh->prepare ( $query )
		or bad ( 120 , "prepare($query) failed: " . $DBI::errstr ) ;
	$sth->execute()
		or bad ( 120 , "execute($query) failed: " . $DBI::errstr ) ;

	my @rv = () ;
	while ( my @data = $sth->fetchrow_array() )
	{
		push ( @rv , @data ) ;
	}

	$sth->finish() ;

	return @rv ;
}

sub sql_do($)
{
	my $query = shift ;

	sql_connect() ;

	$dbh->do ( $query )
		or bad ( 120 , "do($query) failed: " . $DBI::errstr ) ;
}

sub sql_quote($)
{
	sql_connect() ;
	return $dbh->quote ( $_[0] ) ;
}

###############################################################################
#
# Log to file functions.
#
# NOTE: logit() sends to STDOUT. This is normal for a progam running under
# daemontools (as qmail-smtpd normally does.) If you are converting this to
# work with some other MTA, you will need to edit the logit() function in
# order to send the logs wherever you want/need them.

sub logit($)
{
	my $msg = shift ;

	my $name = $0 ;
	$name =~ s|^.*[/\\]|| ;

	my $ppid = getppid() ;

	print "$name" , "[$$][$ppid]" , ": $msg\n" ;
}

sub bad($$)
{
	logit $_[1] ;
	exit $_[0] ;
}

###############################################################################
#
# Log to SQL table function. This is called once for every message, when a
# decision is made to allow or reject the message.

sub log_sql($)
{
	my $disp = shift ;

	logit ( join ( " " , $sender , $recipient , $ip , $disp ) ) ;

	my $qd = sql_quote ( $disp	) ;
	my $qs = sql_quote ( $sender	) ;
	my $qr = sql_quote ( $recipient	) ;
	my $qi = sql_quote ( $ip	) ;

	my $qruleid = ( $ruleid || 0 ) ;

	sql_do <<EOF ;
INSERT INTO log ( disp , sender , recipient , ip , ruleid )
VALUES ( $qd , $qs , $qr , $qi , $qruleid )
EOF
}

###############################################################################
#
# Log processing of individual rules

sub log_process($;$)
{
	return unless $log_rules ;

	my $rtype =   shift ;
	my $rdata = ( shift || '' ) ;

	logit "$phase/$seq($ruleid) $rtype $rdata" ;
}

###############################################################################
#
# Send an error message back to qmail-smtpd using FD4 (only if the USE_FD4
# environment variable exists.) This message ends up being used as the SMTP
# response to the RCPT command which triggered this script's execution.

sub qmail_smtpd_msg($)
{
	# my modified RCPTCHECK code sets this envar
	# if present, a custom error message can be sent back to qmail-smtpd
	unless ( exists $ENV{"USE_FD4"} )
	{
		logit "USE_FD4 is not set, would have sent \"$_[0]\"" ;
		return ;
	}

	open ( M , ">&4" )
		or bad ( 120 , "Can't dup fd 4: $!" ) ;
	print M $_[0] ;
	close M ;
}

###############################################################################
###############################################################################
#
# Refuse the message, either permanently or temporarily. The error message
# (if any) will be sent to qmail-smtpd for use as the SMTP response to the
# RCPT command.
#
# NOTE: These functions DO NOT return, they exit().

sub refuse_with_msg(;$)
{
	my $msg = ( shift || '550 message refused' ) ;
	qmail_smtpd_msg ( $msg ) ;
	exit 100 ;
}

sub delay_with_msg(;$)
{
	my $msg = ( shift || '450 try again later' ) ;
	qmail_smtpd_msg ( $msg ) ;
	exit 111 ;
}

###############################################################################
#
# Convert an IP address to a 32-bit binary number. Returns a binary number or
# undef if not valid.

sub ip2bin($)
{
	unless ( $_[0] =~ m|^\s*(\d+)\.(\d+)\.(\d+)\.(\d+)\s*$| )
	{
		return undef ;
	}

	my ( $a , $b , $c , $d ) = ( $1 , $2 , $3 , $4 ) ;

	if ( $a > 255 || $b > 255 || $c > 255 || $d > 255 )
	{
		return undef ;
	}

	return ( ( $a << 24 ) | ( $b << 16 ) | ( $c << 8 ) | $d ) ;
}

###############################################################################
#
# Convert a subnet mask to a 32-bit binary number. Accepts either CIDR or
# dotted-quad notation (i.e. "24" or "255.255.255.0".) Returns a binary number
# or undef if not valid.

sub mask2bin($)
{
	if ( $_[0] =~ m|^\s*(\d+)\s*$| )
	{
		my $n = $1 ;
		if ( $n > 32 )
		{
			return undef ;
		}

		return ( $n ? 0xffffffff << ( 32 - $n ) : 0 ) ;
	}

	return ip2bin ( $_[0] ) ;
}

###############################################################################
###############################################################################
#
# Rule type processing functions

########################################
# Process an "all messages" rule.
# Note that this function does not return.

sub process_all($)
{
	my $accept = shift ;

	log_process ( 'A' , 'MATCH' ) ;

	if ( $accept )
	{
		log_sql ( 'OK A' ) ;
		exit 0 ;
	}

	log_sql ( 'REFUSE A' ) ;
	refuse_with_msg() ;
}

###############################################################################
#
# Process an email rule, matching the envelope sender address.

sub process_email($$)
{
	my $match  = shift ;
	my $accept = shift ;

	########################################
	# Compare the sender address to the pattern in the rule.
	# If no match, return.

	unless ( $sender =~ m|$match|i )
	{
		log_process ( 'E' , "\"$match\" no" ) ;
		return ;
	}

	log_process ( 'E' , "\"$match\" MATCH" ) ;

	########################################
	# Do whatever action the rule specifies.

	if ( $accept )
	{
		log_sql ( 'OK E' ) ;
		exit 0 ;
	}

	log_sql ( 'REFUSE E' ) ;
	refuse_with_msg() ;
}

###############################################################################
#
# Process an IP rule, matching the SMTP client's IP address.

sub process_ip($$)
{
	my $match  = shift ;
	my $accept = shift ;

	########################################
	# Parse the rule IP/MASK and sender's IP into binary.

	my $bbin = undef ;
	my $tbin = undef ;
	my $mbin = undef ;

	if ( $match =~ m|^([\d\.]+)/([\d\.]+)$| )
	{
		my ( $ipstr , $mstr ) = ( $1 , $2 ) ;
		$bbin = ip2bin   ( $ipstr ) ;
		$mbin = mask2bin ( $mstr  ) ;
	}
	elsif ( $match =~ m|^([\d\.]+)$| )
	{
		$bbin = ip2bin ( $1 ) ;
		$mbin = 0xffffffff ;
	}

	unless ( defined $bbin )
	{
		logit "process_ip(): invalid block \"$match\" (block)" ;
	}

	unless ( defined $mbin )
	{
		logit "process_ip(): invalid block \"$match\" (mask)" ;
	}

	unless ( defined $bbin && defined $mbin )
	{
		delay_with_msg ( '450 internal error' ) ;
	}

	$tbin = ip2bin ( $ip ) ;
	unless ( defined $tbin )
	{
		logit "process_ip(): invalid target \"$ip\"" ;
		delay_with_msg ( '450 internal error' ) ;
	}

	########################################
	# Do the comparison. If no match, return.

	unless ( ( $tbin & $mbin ) == ( $bbin & $mbin ) )
	{
		log_process ( 'I' , "$match no" ) ;
		return ;
	}

	log_process ( 'I' , "$match MATCH" ) ;

	########################################
	# Do whatever action the rule specifies.

	if ( $accept )
	{
		log_sql ( 'OK I' ) ;
		exit 0 ;
	}

	log_sql ( 'REFUSE I' ) ;
	refuse_with_msg ( '550 message refused' ) ;
}

###############################################################################
#
# Process an RBL rule, checking the SMTP client's IP address against an RBL.

sub process_rbl($$)
{
	my $zone   = shift ;
	my $accept = shift ;

	########################################
	# Build the name we will be searching for.

	unless ( $ip =~ m|^(\d+)\.(\d+)\.(\d+)\.(\d+)$| )
	{
		logit "process_rbl(): invalid IP \"$ip\"" ;
		delay_with_msg ( '450 internal error' ) ;
	}

	my $name = "$4.$3.$2.$1.$zone" ;

	########################################
	# Search for TXT records, then A records.

	my $data = undef ;

	my $res = new Net::DNS::Resolver ;

	my $q = $res->search ( $name , 'TXT' ) ;
	if ( $q )
	{
		for my $rr ( $q->answer() )
		{
				$data = $rr->txtdata() ;
		}
	}

	unless ( $data )
	{
		$q = $res->search ( $name , 'A' ) ;
		if ( $q )
		{
			for my $rr ( $q->answer() )
			{
				$data = 'A' ;
			}
		}
	}

	########################################
	# If we didn't find anything, return.

	unless ( $data )
	{
		log_process ( 'R' , "$zone no" ) ;
		return ;
	}

	log_process ( 'R' , "$zone MATCH" ) ;

	########################################
	# Do whatever action the rule specifies.

	if ( $accept )
	{
		log_sql ( 'OK R' ) ;
		exit 0 ;
	}

	log_sql ( 'REFUSE R' ) ;

	if ( $data eq 'A' )
	{
		refuse_with_msg ( "550 refused, RBL $zone" ) ;
	}
	else
	{
		refuse_with_msg ( "550 $data" ) ;
	}
}

###############################################################################
#
# Process a greylist rule.
# Note that this function does not return.

sub process_greylist($$)
{
	my $delay  = shift ;
	my $accept = shift ;

	my $now = time() ;

	########################################
	# Build the query.

	my $qsender    = sql_quote ( $sender    ) ;
	my $qrecipient = sql_quote ( $recipient ) ;
	my $qip        = sql_quote ( $ip        ) ;

	my $query = <<EOF ;
SELECT id , first FROM greylist
WHERE $qsender LIKE sender
AND $qrecipient LIKE recipient
AND $qip LIKE ip
EOF

	########################################
	# Run the query.

	my ( $gid , $first ) = sql_select ( $query ) ;

	########################################
	# Did we find anything? If not, add an entry for it.

	unless ( defined $first )
	{
		log_process ( 'G' , "$delay FIRST" ) ;
		log_sql ( 'GREY1' ) ;

		if ( $grey_match_24 )
		{
			$ip =~ m|^(\d+\.\d+\.\d+)| ;
			$qip = sql_quote ( "$1.\%" ) ;
		}

		$query = <<EOF ;
INSERT INTO greylist ( sender , recipient , ip , first , last )
VALUES ( $qsender , $qrecipient , $qip , $now , $now )
EOF
		sql_do ( $query ) ;

		delay_with_msg ( "450 greylisted, wait" ) ;
	}

	########################################
	# see if it's been long enough to accept the message.

	my $when_ok = $first + $delay ;
	if ( time() < $when_ok )
	{
		log_process ( 'G' , "$delay TOO SOON" ) ;
		log_sql ( 'GREY2' ) ;

		sql_do <<EOF ;
UPDATE greylist SET count_pre = count_pre + 1 , last = $now
WHERE id = $gid
EOF

		delay_with_msg ( "450 still greylisted, keep waiting" ) ;
	}

	########################################
	# good to go

	log_process ( 'G' , "$delay OK" ) ;
	log_sql ( 'OK G' ) ;

	sql_do <<EOF ;
UPDATE greylist SET count_msg = count_msg + 1 , last = $now
WHERE id = $gid
EOF

	exit 0 ;
}

###############################################################################
#
# Process a target rule, matching the recipient address.

sub process_target($$)
{
	my $match  = shift ;
	my $accept = shift ;

	########################################
	# Compare the recipient address to the pattern in the rule.
	# If no match, return.

	unless ( $original_recipient =~ m|$match|i )
	{
		log_process ( 'T' , "\"$match\" no" ) ;
		return ;
	}

	log_process ( 'T' , "\"$match\" MATCH" ) ;

	########################################
	# Do whatever action the rule specifies.

	if ( $accept )
	{
		log_sql ( 'OK T' ) ;
		exit 0 ;
	}

	log_sql ( 'REFUSE T' ) ;
	refuse_with_msg() ;
}

###############################################################################
#
# Process an AUTH rule

sub process_auth($)
{
	my $accept = shift ;

	########################################
	# Does the SMTP_AUTH_USER variable exist?
	# If not, return.

	my $auth_user = ( $ENV{'SMTP_AUTH_USER'} || '' ) ;
	unless ( $auth_user )
	{
		log_process ( 'U' , "no" ) ;
		return ;
	}

	log_process ( 'U' , "MATCH ($auth_user)" ) ;

	########################################
	# Do whatever action the rule specifies.

	if ( $accept )
	{
		log_sql ( 'OK U' ) ;
		exit 0 ;
	}

	log_sql ( 'REFUSE U' ) ;
	refuse_with_msg() ;
}

###############################################################################
###############################################################################
###############################################################################
#
# MAIN PROCESSING STARTS HERE

if ( -f '/var/qmail/control/g4dump' )
{
	map { logit "DUMP: $_=\"$ENV{$_}\"" } sort keys %ENV ;
}

###############################################################################
#
# Read configuration

$sql_server	= ( $ENV{'G4DBCONN'} || bad ( 120 , 'G4DBCONN not set' ) ) ;
$sql_userid	= ( $ENV{'G4DBUSER'} || '' ) ;
$sql_passwd	= ( $ENV{'G4DBPASS'} || '' ) ;
$grey_match_24	= ( $ENV{'G4GREY24'} || 0  ) ;

########################################
# Read message information (sender, recipient, sending IP)

$sender    = lc ( $ENV{"SENDER"}      || "" ) ;
$recipient = lc ( $ENV{"RECIPIENT"}   || bad ( 120, "RECIPIENT not set"   ) ) ;
$ip        =    ( $ENV{"TCPREMOTEIP"} || bad ( 120, "TCPREMOTEIP not set" ) ) ;

########################################
# If there is no sender, it's a bounce.

unless ( $sender )
{
	$sender = '<>' ;
	log_sql ( 'RCV BOUNCE' ) ;
	exit 0 ;
}

########################################
# Clean up the recipient address.
# Convert "user-xyz@domain" to "user@domain".

$original_recipient = $recipient ;

my ( $zuser , $zdom ) = split ( m|\@| , $recipient , 2 ) ;
$zuser =~ s|\-.*|| ;
$recipient = "$zuser\@$zdom" ;

########################################
# Query rules from database

my $qsender	= sql_quote ( $sender    ) ;
my $qrecipient	= sql_quote ( $recipient ) ;
my $qip		= sql_quote ( $ip        ) ;

########################################
# Search for debugging rules

my $query = <<EOF ;
SELECT COUNT(*)
FROM rules
WHERE $qrecipient LIKE recipient
AND type = 'D'
EOF

( $log_rules ) = sql_select ( $query ) ;

########################################
# Search for mail processing rules

$query = <<EOF ;
SELECT id , phase , seq , accept , type , sender , ip , rbl , delay , target
FROM rules
WHERE $qrecipient LIKE recipient
AND type != 'D'
ORDER BY phase , seq
EOF

my @data = sql_select ( $query ) ;

########################################
# Process rules

while ( $#data > -1 )
{
	########################################
	# Read data for this rule

	$ruleid		= ( shift @data || 0  ) ;
	$phase		= ( shift @data || 0  ) ;
	$seq		= ( shift @data || 0  ) ;
	my $r_accept	= ( shift @data || 0  ) ;
	my $r_type	= ( shift @data || '' ) ;
	my $r_sender	= ( shift @data || '' ) ;
	my $r_ip	= ( shift @data || '' ) ;
	my $r_rbl	= ( shift @data || '' ) ;
	my $r_delay	= ( shift @data || 0  ) ;
	my $r_target	= ( shift @data || '' ) ;

	########################################
	# Clean up data

	$r_ip =~ s|^\s*|| ;
	$r_ip =~ s|\s*$|| ;

	########################################
	# Process the rule

	if ( $r_type eq 'A' )
	{
		process_all ( $r_accept ) ;
	}
	elsif ( $r_type eq 'E' )
	{
		process_email ( $r_sender , $r_accept ) ;
	}
	elsif ( $r_type eq 'I' )
	{
		process_ip ( $r_ip , $r_accept ) ;
	}
	elsif ( $r_type eq 'R' )
	{
		process_rbl ( $r_rbl , $r_accept ) ;
	}
	elsif ( $r_type eq 'T' )
	{
		process_target ( $r_target , $r_accept ) ;
	}
	elsif ( $r_type eq 'U' )
	{
		process_auth ( $r_accept ) ;
	}
	elsif ( $r_type eq 'G' )
	{
		process_greylist ( $r_delay , $r_accept ) ;
	}
	else
	{
		logit "$phase/$seq($ruleid) unknown rule type \"$r_type\"" ;
		delay_with_msg ( '450 internal error' ) ;
	}
}

########################################
# if we reach this point, no rules matched.
# for safety, the default is to accept the message.

$ruleid = 0 ;
log_sql ( 'OK' ) ;
exit 0 ;
